mod sieve;

use std::{
    collections::HashMap,
    env,
    sync::Arc,
    thread::{self, available_parallelism},
    time::Instant,
};

use itertools::Itertools as _;
use sieve::get_primes;

const MAX_SIEVED_PRIMES: usize = 100_000_000;

fn count_prime_factor_exponent(prime: u64, mut nr: u64) -> u64 {
    let mut count = 0;

    while nr % prime == 0 {
        nr /= prime;
        count += 1;
    }

    count
}

fn prime_factors(nr: u64, primes: &[u64]) -> u64 {
    let mut num_teilers = 1;

    // safety: nr <= 2.pow(53) must be fulfilled,
    //         otherwise we have a precision loss
    assert!(nr <= 2u64.pow(53));
    #[allow(
        clippy::cast_sign_loss,
        clippy::cast_possible_truncation,
        clippy::cast_precision_loss
    )]
    let nr_sqrt = (nr as f64).sqrt().ceil() as u64;

    for prime in primes {
        if *prime > nr_sqrt {
            break;
        }

        let count = count_prime_factor_exponent(*prime, nr);
        if count > 0 {
            num_teilers *= count + 1;
        }
    }

    num_teilers
}

fn the_thread(mut start: u64, end: u64, primes: &[u64]) -> HashMap<u64, u64> {
    let mut max_teilers = 0;
    let mut max_teilers_hashmap = HashMap::new();

    let mut step_size = 10;

    if start == 0 {
        start = 2;
        step_size = 2;
    }

    for nr in (start..=end).step_by(step_size) {
        let teilers = prime_factors(nr, primes);
        if teilers > max_teilers {
            max_teilers = teilers;
            max_teilers_hashmap.insert(nr, teilers);
        }
    }

    max_teilers_hashmap
}

fn calculate_chunk_bounds(i: usize, num_threads: usize, max_nr: usize) -> (usize, usize) {
    let chunk_size = max_nr / num_threads;
    let start = i * chunk_size;
    let end = ((i + 1) * chunk_size).min(max_nr);
    (start, end)
}

fn main() {
    let max_nr = env::args()
        .nth(1)
        .expect("Usage: hcn <max_nr>")
        .parse::<usize>()
        .expect("Invalid max_nr");

    let start = Instant::now();
    println!("Precalculating primes...");
    let primes = get_primes(MAX_SIEVED_PRIMES);
    println!("{} primes. Took {:?}", primes.len(), start.elapsed());

    let primes = Arc::new(primes);

    #[allow(clippy::unwrap_used)]
    let num_threads = available_parallelism().unwrap().get();
    let mut threads = Vec::with_capacity(num_threads);

    let now = Instant::now();
    for i in 0..num_threads {
        let (start, end) = calculate_chunk_bounds(i, num_threads, max_nr);
        let primes = primes.clone();
        threads.push(thread::spawn(move || {
            the_thread(start as u64, end as u64, &primes)
        }));
    }

    let mut results = HashMap::new();
    for thread in threads {
        let thread_result = thread.join().expect("Thread failed to join (panicked?)");
        results.extend(thread_result);
    }

    let mut max_teilers = 0;

    for (nr, teilers) in results.into_iter().sorted() {
        if teilers > max_teilers {
            println!("{nr}: {teilers}");
            max_teilers = teilers;
        }
    }

    println!("Took {:?}", now.elapsed());
}
